#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <sys/time.h>
#include <assert.h>
#include "conv_weight.h"

#define BLOCKDIM_Y 16
#define BLOCKDIM_X 16

const double kMicro = 1.0e-6;
#define FLOPS (21*125+27) //342?
#define checkMalloc(x) if (!x) { fprintf(stderr, "Malloc failed\n"); return 0; }

//256
int outputWidth = 256;
//256
int outputHeight = 256;
//256
int outputDepth = 256;
char fileName[512UL] = "Engine.raw";

/**
 * Allocates input (data) and output (cornerness) array.
 * Both arrays are 2D array of size height (= # of rows) x width (= # of columns). However, memory
 * is allocated as a 1D array and 2D array pointers are initialized so that 2D indexing finds
 * correct memory locations in the 1D memory region.
 *
 * @return 1 if no error, 0 otherwise.
 */

float ***allocateArrays3D(int k,int m,int n)
{
  float ***E = (float ***)(((void *)0));
  int nx = n;
  int ny = m;
  int nk = k;
  E = ((float ***)(malloc((((sizeof(float **)) * (nk))))));
  (E != (0))?((void )0) : __assert_fail(("E"),("HarrisPoint3D_5x5x5.c"),(100),("float ***allocateArrays3D(int, int, int)"));
  E[0] = ((float **)(malloc(((((sizeof(float *)) * (nk)) * (ny))))));
  (E[0])[0] = ((float *)(malloc((((((sizeof(float )) * (nx)) * (ny)) * (nk))))));
  int jj;
  int kk;
  for (kk = 0; kk < nk; kk++) {
    if (kk > 0) {
      E[kk] = ((E[kk - 1]) + ny);
      (E[kk])[0] = (((E[kk - 1])[0]) + (ny * nx));
    }
    for (jj = 1; jj < ny; jj++) {
      (E[kk])[jj] = (((E[kk])[jj - 1]) + nx);
    }
  }
  return E;
}

void free3D(float ***E,int depth)
{
  free((((E[0])[0])));
//for (int i = 0; i < depth; i++) {
//	free(E[i]);
//}
  free(((E[0])));
  free((E));
}
//calculate l2norm for comparison

void calculatel2Norm(float ***E,int n,int m,int k,int nIters)
{
  int i;
  int j;
  int l = 0;
  float mx = ((-1));
  float l2norm = (0);
  for (l = 3; l < (k - 3); l++) {
    for (j = 3; j < (m - 3); j++) {
      for (i = 3; i < (n - 3); i++) {
        l2norm += ((((E[l])[j])[i]) * (((E[l])[j])[i]));
        if ((((E[l])[j])[i]) > mx) 
          mx = (((E[l])[j])[i]);
      }
    }
  }
  l2norm /= ((float )(m * (n * k)));
  l2norm = ((sqrt((l2norm))));
  printf((":N %d M %d K %d, iteration %d\n"),n,m,k,nIters);
  printf((":max: %20.12e, l2norm: %20.12e\n"),(mx),(l2norm));
}

double getTime()
{
  struct timeval TV;
  const int RC = gettimeofday(&TV,0);
  if (RC == (-1)) {
    printf(("ERROR: Bad call to gettimeofday\n"));
    return ((-1));
  }
  return ((double )(TV.tv_sec)) + (kMicro * ((double )(TV.tv_usec)));
// end getTime()
}
/**
 * Returns determinant of 3 x 3 matrix, [Lx Ly Lz]' * [Lx Ly Lz]
 */


int loadData3D(float ***data)
{
  int rc;
  int j;
  int k;
  FILE *fp;
  float rowBuf[outputWidth];
  if ((fp = fopen((fileName),("r"))) == ((FILE *)(((void *)0)))) {
    fprintf(stderr,("File not found: %s\n"),fileName);
    return 0;
  }
// read one row and copy into inner region.
  for (k = 0; k < outputDepth; ++k) {
    for (j = 0; j < outputHeight; ++j) {
      if ((rc = ((fread((rowBuf),((sizeof(float ))),(outputWidth),fp)))) != outputWidth) {
        fprintf(stderr,("File read error: (%d, %d): %d\n"),k,j,rc);
        return 0;
      }
      memcpy(((((data[k + 3])[j + 3]) + 3)),(rowBuf),(((sizeof(float )) * (outputWidth))));
    }
  }
  fclose(fp);
  return 1;
}
void zeroOutArray(float ***data,int d,int h,int w)
{
  for (int i = 0; i < d; ++i) {
    for (int j = 0; j < h; ++j) {
      for (int k = 0; k < w; ++k) {
        ((data[i])[j])[k] = 0.0f;
//if(i==5)
//data[i][j][k] = 1.0f;
      }
    }
  }
}
__global__ void test_1_1527(cudaPitchedPtr dev_3_w,cudaPitchedPtr dev_1_data,cudaPitchedPtr dev_2_cornerness,int depth,int height,int width,int num2blockDim_1_1527,float invYnumblockDim_1_1527);

int main(int argc,char **argv)
{
  int rc;
  float ***data = (float ***)(((void *)0));
  float ***cornerness = (float ***)(((void *)0));
  int depth;
  int height;
  int width;
  depth = (outputDepth + 6);
  height = (outputHeight + 6);
  width = (outputWidth + 6);
// get arguments
/*
  if ((rc = parseArgs3D(argc,argv,&depth,&height,&width)) == 0) {
    return 0;
    }*/
// allocate arrays
  if ((data = allocateArrays3D(depth,height,width)) == ((float ***)(((void *)0)))) {
    fprintf(stderr,("Data array allocation failed\n"));
    return 0;
  }
  if ((cornerness = allocateArrays3D(depth,height,width)) == ((float ***)(((void *)0)))) {
    fprintf(stderr,("Correctness array allocation failed\n"));
    return 0;
  }
  printf(("Memory is successfully allocated.\n"));
  zeroOutArray(data,depth,height,width);
  zeroOutArray(cornerness,depth,height,width);
// load data
  if (!((rc = loadData3D(data)) != 0)) {
    fprintf(stderr,("Data load error: %s\n"),fileName);
    return 0;
  }
  printf(("Data is successfully loaded.\n"));

  cudaError_t stat_dev_1_data;
  cudaExtent ext_dev_1_data = make_cudaExtent(width * sizeof(float ),height,depth);

  cudaPitchedPtr dev_1_data;
  stat_dev_1_data = cudaMalloc3D(&dev_1_data,ext_dev_1_data);
  if (stat_dev_1_data != cudaSuccess) 
    fprintf(stderr,"%s\n",cudaGetErrorString(stat_dev_1_data));

  cudaMemcpy3DParms param_dev_1_data = {0};
  param_dev_1_data.srcPtr = make_cudaPitchedPtr(((void *)(data[0][0])),width * sizeof(float ),width,height);
  param_dev_1_data.dstPtr = dev_1_data;
  param_dev_1_data.extent = ext_dev_1_data;
  param_dev_1_data.kind = cudaMemcpyHostToDevice;
  stat_dev_1_data = cudaMemcpy3D(&param_dev_1_data);
  if (stat_dev_1_data != cudaSuccess) 
    fprintf(stderr,"%s\n",cudaGetErrorString(stat_dev_1_data));

  cudaError_t stat_dev_2_cornerness;
  cudaExtent ext_dev_2_cornerness = make_cudaExtent(width * sizeof(float ),height,depth);

  cudaPitchedPtr dev_2_cornerness;
  stat_dev_2_cornerness = cudaMalloc3D(&dev_2_cornerness,ext_dev_2_cornerness);
  if (stat_dev_2_cornerness != cudaSuccess) 
    fprintf(stderr,"%s\n",cudaGetErrorString(stat_dev_2_cornerness));

  cudaMemcpy3DParms param_dev_2_cornerness = {0};
  param_dev_2_cornerness.srcPtr = make_cudaPitchedPtr(((void *)(cornerness[0][0])),width * sizeof(float ),width,height);
  param_dev_2_cornerness.dstPtr = dev_2_cornerness;
  param_dev_2_cornerness.extent = ext_dev_2_cornerness;
  param_dev_2_cornerness.kind = cudaMemcpyHostToDevice;
  stat_dev_2_cornerness = cudaMemcpy3D(&param_dev_2_cornerness);
  if (stat_dev_2_cornerness != cudaSuccess) 
    fprintf(stderr,"%s\n",cudaGetErrorString(stat_dev_2_cornerness));

  cudaError_t stat_dev_3_w;
  cudaExtent ext_dev_3_w = make_cudaExtent(5 * sizeof(float ),5,5);

  cudaPitchedPtr dev_3_w;
  stat_dev_3_w = cudaMalloc3D(&dev_3_w,ext_dev_3_w);
  if (stat_dev_3_w != cudaSuccess) 
    fprintf(stderr,"%s\n",cudaGetErrorString(stat_dev_3_w));

  cudaMemcpy3DParms param_dev_3_w = {0};
  param_dev_3_w.srcPtr = make_cudaPitchedPtr(((void *)(w[0][0])),5 * sizeof(float ),5,5);
  param_dev_3_w.dstPtr = dev_3_w;
  param_dev_3_w.extent = ext_dev_3_w;
  param_dev_3_w.kind = cudaMemcpyHostToDevice;
  stat_dev_3_w = cudaMemcpy3D(&param_dev_3_w);
  if (stat_dev_3_w != cudaSuccess) 
    fprintf(stderr,"%s\n",cudaGetErrorString(stat_dev_3_w));
  double time_elapsed = getTime();
  double Gflops = 0.0;
{
// main loop

    int num3blockDim_1_1527 = ((depth - 3) - 3) % 1 == 0?((depth - 3) - 3) / 1 : ((depth - 3) - 3) / 1 + 1;
    int num2blockDim_1_1527 = ((height - 3) - 3) % 16 == 0?((height - 3) - 3) / 16 : ((height - 3) - 3) / 16 + 1;
    int num1blockDim_1_1527 = ((width - 3) - 3) % 16 == 0?((width - 3) - 3) / 16 : ((width - 3) - 3) / 16 + 1;
    float invYnumblockDim_1_1527 = 1.00000F / num2blockDim_1_1527;
    dim3 blockDim_1_1527(16,16,1);
    dim3 gridDim_1_1527(num1blockDim_1_1527,num2blockDim_1_1527*num3blockDim_1_1527);
    test_1_1527<<<gridDim_1_1527,blockDim_1_1527>>>(dev_3_w,dev_1_data,dev_2_cornerness,depth,height,width,num2blockDim_1_1527,invYnumblockDim_1_1527);
    cudaThreadSynchronize();
    cudaError_t err_test_1_1527 = cudaGetLastError();
    if (err_test_1_1527) 
      fprintf(stderr,"%s\n",cudaGetErrorString(err_test_1_1527));
//end of parallel region 
  }
  time_elapsed = (getTime() - time_elapsed);
  printf(("Computation done\n"));

  cudaMemcpy3DParms param2_dev_2_cornerness = {0};
  param2_dev_2_cornerness.srcPtr = dev_2_cornerness;
  param2_dev_2_cornerness.dstPtr = make_cudaPitchedPtr(((void *)(cornerness[0][0])),width * sizeof(float ),width,height);
  param2_dev_2_cornerness.extent = ext_dev_2_cornerness;
  param2_dev_2_cornerness.kind = cudaMemcpyDeviceToHost;
  stat_dev_2_cornerness = cudaMemcpy3D(&param2_dev_2_cornerness);
  if (stat_dev_2_cornerness != cudaSuccess) 
    fprintf(stderr,"%s\n",cudaGetErrorString(stat_dev_2_cornerness));
  calculatel2Norm(cornerness,width,height,depth,1);
  Gflops = (((((((width - 6) * (height - 6)) * (depth - 6))) * 1.0e-9) * (((21 * 125) + 27))) / time_elapsed);
  printf(("\nProgram  \tTime(sec)\tGflops\n"));
  printf(("==========================================\n"));
  printf(("%s%3.3f \t%5.3f\n"),"Point3D   ",time_elapsed,Gflops);

  free3D(data,depth);
  free3D(cornerness,depth);
  printf(("Done\n"));
  cudaFree(dev_3_w.ptr);
  cudaFree(dev_1_data.ptr);
  cudaFree(dev_2_cornerness.ptr);
  return 0;
}

__global__ void test_1_1527(cudaPitchedPtr dev_3_w,cudaPitchedPtr dev_1_data,cudaPitchedPtr dev_2_cornerness,int depth,int height,int width,int num2blockDim_1_1527,float invYnumblockDim_1_1527)
{

  float *w = (float *)(dev_3_w.ptr);

  float *data = (float *)(dev_1_data.ptr);
  int widthdata = dev_1_data.pitch / sizeof(float );
  int slicedata = dev_1_data.ysize * widthdata;

  float *cornerness = (float *)(dev_2_cornerness.ptr);
  int widthcornerness = dev_2_cornerness.pitch / sizeof(float );
  int slicecornerness = dev_2_cornerness.ysize * widthcornerness;

  float blocksInY = num2blockDim_1_1527;
  float invBlocksInY = invYnumblockDim_1_1527;

    int _upperb_y = (height - 3) - 1;
    int _upperb_x = (width - 3) - 1;
    int _idx = threadIdx.x + 3;
    int _gidx = _idx + blockDim.x * blockIdx.x;
    int _idy = threadIdx.y + 3;
    int _gidy = _idy + blockDim.y * 1 * blockIdx.y;
    int _idz = threadIdx.z + 3;
    int blockIdxz = blockIdx.y * invBlocksInY;
    int blockIdxy = blockIdx.y - blockIdxz * blocksInY;
    _gidy = _idy + blockIdxy * blockDim.y;
    int _gidz = _idz + blockIdxz * blockDim.z;
    int indexcornerness = _gidx + _gidy * widthcornerness + _gidz * slicecornerness;
    int indexdata = _gidx + _gidy * widthdata + _gidz * slicedata;

      if (_gidz >= 3 && _gidz <= (depth - 3) - 1) {
            if (_gidy >= 3 && _gidy <= (height - 3) - 1) {
                  if (_gidx >= 3 && _gidx <= (width - 3) - 1) {

		      float dz = data[indexdata-widthdata*2]; 

                      cornerness[indexcornerness] = dz*  w[0];

		  }
	    }
      }
}
